__all__ = ["MMDetModelAdapter"]

from icevision.imports import *
from icevision.utils import *
from icevision.metrics import *
from icevision.engines.lightning.lightning_model_adapter import LightningModelAdapter

from icevision.models.mmdet.common.bbox import convert_raw_predictions


class MMDetModelAdapter(LightningModelAdapter, ABC):
    """Lightning module specialized for EfficientDet, with metrics support.

    The methods `forward`, `training_step`, `validation_step`, `validation_epoch_end`
    are already overriden.

    # Arguments
        model: The pytorch model to use.
        metrics: `Sequence` of metrics to use.

    # Returns
        A `LightningModule`.
    """

    def __init__(self, model: nn.Module, metrics: List[Metric] = None):
        super().__init__(metrics=metrics)
        self.model = model

    @abstractmethod
    def convert_raw_predictions(self, batch, raw_preds, records):
        """Convert raw predictions from the model to library standard."""

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        data, samples = batch

        outputs = self.model.train_step(data=data, optimizer=None)

        for k, v in outputs["log_vars"].items():
            self.log(f"train/{k}", v)

        return outputs["loss"]

    def validation_step(self, batch, batch_idx):
        data, records = batch
        self.model.eval()
        with torch.no_grad():
            outputs = self.model.train_step(data=data, optimizer=None)
            raw_preds = self.model.forward_test(
                imgs=[data["img"]], img_metas=[data["img_metas"]]
            )

        preds = self.convert_raw_predictions(
            batch=data, raw_preds=raw_preds, records=records
        )
        self.accumulate_metrics(preds)

        for k, v in outputs["log_vars"].items():
            self.log(f"valid/{k}", v)

        # TODO: is train and eval model automatically set by lighnting?
        self.model.train()

    def validation_epoch_end(self, outs):
        self.finalize_metrics()



__all__ = ["process_bbox_predictions", "_end2end_detect", "draw_img_and_boxes"]

from icevision.imports import *
from icevision.core import *
from icevision.data import *
from icevision.tfms.albumentations.albumentations_helpers import (
    get_size_without_padding,
)
from icevision.tfms.albumentations import albumentations_adapter

from icevision.utils.imageio import *
from icevision.visualize.draw_data import *
from icevision.visualize.utils import *

DEFAULT_FONT_PATH = get_default_font()


def _end2end_detect(
    img: Union[PIL.Image.Image, Path, str],
    transforms: albumentations_adapter.Adapter,
    model: torch.nn.Module,
    class_map: ClassMap,
    detection_threshold: float = 0.5,
    predict_fn: Callable = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_score: bool = True,
    font_path: Optional[os.PathLike] = DEFAULT_FONT_PATH,
    font_size: Union[int, float] = 12,
    label_color: Union[np.array, list, tuple, str] = ("#C4C4C4"),  # Mild Gray
    return_as_pil_img=True,
    return_img=True,
):
    """
    Run Object Detection inference (only `bboxes`) on a single image.

    Parameters
    ----------
    img: image to run inference on. It can be a string, Path or PIL.Image
    transforms: icevision albumentations transforms
    model: model to run inference with
    class_map: ClassMap with the available categories
    detection_threshold: confidence threshold below which bounding boxes are discarded
    display_label: display or not a bounding box label(i.e class)
    display_bbox: display or not a bounding box
    display_score: display or not a bounding box score
    font_path: path to the font file used
    font_size: font size
    label_color: a <collection> of RGB values or a hex code string that defines
                   the color of all the plotted labels
    return_as_pil_img: if True a PIL image is returned otherwise a numpy array is returned
    return_img: whether we should also return an image in addition to the bounding boxes, labels, and scores

    Returns
    -------
    A dictionnary with categories, scores, bounding box coordinates, image height and width,
                   and optionally a PIL Image or a numpy array (image).
                   Bounding boxes are adjusted to the original image size and aspect ratio
    """
    if isinstance(img, (str, Path)):
        img = PIL.Image.open(Path(img))

    infer_ds = Dataset.from_images([np.array(img)], transforms, class_map=class_map)
    pred = predict_fn(model, infer_ds, detection_threshold=detection_threshold)[0]
    pred = process_bbox_predictions(pred, img, transforms.tfms_list)
    record = pred.pred

    # draw image, return it as PIL image by default otherwise as numpy array
    if return_img:
        pred_img = draw_record(
            record=pred,
            class_map=class_map,
            display_label=display_label,
            display_score=display_score,
            display_bbox=display_bbox,
            font_path=font_path,
            font_size=font_size,
            label_color=label_color,
            return_as_pil_img=return_as_pil_img,
        )
        record.set_img(pred_img)
    else:
        record._unload()

    w, h = img.shape
    record.set_img_size(ImgSize(width=w, height=h))

    pred_dict = record.as_dict()

    # expose img at the root instead of having under the `common` key
    if return_img:
        pred_dict["img"] = pred_img
    else:
        pred_dict["img"] = None

    pred_dict["width"] = w
    pred_dict["height"] = h
    # delete the `common` key that holds both the `img` and its shape
    del pred_dict["common"]

    # return a dict that contains the image with its predicted boxes (i.e. with resized boxes that match the original image size)
    # labels, and prediction scores
    return pred_dict


def process_bbox_predictions(
    pred: Prediction,
    img: PIL.Image.Image,
    transforms: List[Any],
) -> List[Dict[str, Any]]:
    """
    Postprocess prediction.

    Parameters
    ----------
    pred: icevision prediction object
    img: original image, before any model-pre-processing done
    transforms: list of model-pre-processing transforms

    Returns
    -------
    List of dicts with class, score and bbox coordinates
    """
    bboxes = []
    for bbox, score, label in zip(
        pred.pred.detection.bboxes,
        pred.pred.detection.scores,
        pred.pred.detection.labels,
    ):
        xmin, ymin, xmax, ymax = postprocess_bbox(
            img, bbox, transforms, pred.pred.height, pred.pred.width
        )

        bbox = BBox.from_xyxy(xmin, ymin, xmax, ymax)
        bboxes.append(bbox)

    pred.pred.img = np.array(img)
    pred.pred.detection.set_bboxes(bboxes)
    return pred


def postprocess_bbox(
    img: PIL.Image.Image, bbox: BBox, transforms: List[Any], h_after: int, w_after: int
) -> Tuple[int, int, int, int]:
    """
    Post-process predicted bbox to adjust coordinates to input image size.

    Parameters
    ----------
    img: original image, before any model-pre-processing done
    bbox: predicted bbox
    transforms: list of model-pre-processing transforms
    h_after: height of image after model-pre-processing transforms
    w_after: width of image after model-pre-processing transforms

    Returns
    -------
    Tuple with (xmin, ymin, xmax, ymax) rescaled and re-adjusted to match the original image size
    """
    w_before, h_before = img.size
    h_after, w_after = get_size_without_padding(transforms, img, h_after, w_after)
    pad = np.abs(h_after - w_after) // 2

    h_scale, w_scale = h_after / h_before, w_after / w_before
    if h_after < w_after:
        xmin, xmax, ymin, ymax = (
            int(bbox.xmin),
            int(bbox.xmax),
            int(bbox.ymin) - pad,
            int(bbox.ymax) - pad,
        )
    else:
        xmin, xmax, ymin, ymax = (
            int(bbox.xmin) - pad,
            int(bbox.xmax) - pad,
            int(bbox.ymin),
            int(bbox.ymax),
        )

    xmin, xmax, ymin, ymax = (
        max(xmin, 0),
        min(xmax, w_after),
        max(ymin, 0),
        min(ymax, h_after),
    )
    xmin, xmax, ymin, ymax = (
        int(xmin / w_scale),
        int(xmax / w_scale),
        int(ymin / h_scale),
        int(ymax / h_scale),
    )

    return xmin, ymin, xmax, ymax


def draw_img_and_boxes(
    img: Union[PIL.Image.Image, np.ndarray],
    bboxes: dict,
    class_map,
    display_score: bool = True,
    label_color: Union[np.array, list, tuple, str] = (255, 255, 0),
    label_border_color: Union[np.array, list, tuple, str] = (255, 255, 0),
) -> PIL.Image.Image:

    if not isinstance(img, PIL.Image.Image):
        img = np.array(img)

    # convert dict to record
    record = ObjectDetectionRecord()
    w, h = img.shape
    record.img = np.array(img)
    record.set_img_size(ImgSize(width=w, height=h))
    record.detection.set_class_map(class_map)
    for bbox in bboxes:
        record.detection.add_bboxes([BBox.from_xyxy(*bbox["bbox"])])
        record.detection.add_labels([bbox["class"]])
        record.detection.set_scores(bbox["score"])

    pred_img = draw_sample(
        record,
        display_score=display_score,
        label_color=label_color,
        label_border_color=label_border_color,
    )

    return pred_img



__all__ = [
    "resnet50_caffe_dc5_1x",
    "resnet50_caffe_fpn_1x",
    "resnet50_fpn_1x",
    "resnet50_fpn_2x",
    "resnet101_caffe_fpn_1x",
    "resnet101_fpn_1x",
    "resnet101_fpn_2x",
    "resnext101_32x4d_fpn_1x",
    "resnext101_32x4d_fpn_2x",
    "resnext101_64x4d_fpn_1x",
    "resnext101_64x4d_fpn_2x",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MMDetFasterRCNNBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="faster_rcnn", **kwargs)


base_config_path = mmdet_configs_path / "faster_rcnn"
base_weights_url = "http://download.openmmlab.com/mmdetection/v2.0/faster_rcnn"

resnet50_caffe_dc5_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r50_caffe_dc5_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r50_caffe_dc5_1x_coco/faster_rcnn_r50_caffe_dc5_1x_coco_20201030_151909-531f0f43.pth",
)

resnet50_caffe_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r50_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r50_caffe_fpn_1x_coco/faster_rcnn_r50_caffe_fpn_1x_coco_bbox_mAP-0.378_20200504_180032-c5925ee5.pth",
)

resnet50_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r50_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r50_fpn_1x_coco/faster_rcnn_r50_fpn_1x_coco_20200130-047c8118.pth",
)

resnet50_fpn_2x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r50_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r50_fpn_2x_coco/faster_rcnn_r50_fpn_2x_coco_bbox_mAP-0.384_20200504_210434-a5d8aa15.pth",
)

resnet101_caffe_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r101_caffe_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r101_caffe_fpn_1x_coco/faster_rcnn_r101_caffe_fpn_1x_coco_bbox_mAP-0.398_20200504_180057-b269e9dd.pth",
)

resnet101_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r101_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r101_fpn_1x_coco/faster_rcnn_r101_fpn_1x_coco_20200130-f513f705.pth",
)

resnet101_fpn_2x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_r101_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_r101_fpn_2x_coco/faster_rcnn_r101_fpn_2x_coco_bbox_mAP-0.398_20200504_210455-1d2dac9c.pth",
)

resnext101_32x4d_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_x101_32x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_x101_32x4d_fpn_1x_coco/faster_rcnn_x101_32x4d_fpn_1x_coco_20200203-cff10310.pth",
)

resnext101_32x4d_fpn_2x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_x101_32x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_x101_32x4d_fpn_2x_coco/faster_rcnn_x101_32x4d_fpn_2x_coco_bbox_mAP-0.412_20200506_041400-64a12c0b.pth",
)

resnext101_64x4d_fpn_1x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_x101_64x4d_fpn_1x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_x101_64x4d_fpn_1x_coco/faster_rcnn_x101_64x4d_fpn_1x_coco_20200204-833ee192.pth",
)

resnext101_64x4d_fpn_2x = MMDetFasterRCNNBackboneConfig(
    config_path=base_config_path / "faster_rcnn_x101_64x4d_fpn_2x_coco.py",
    weights_url=f"{base_weights_url}/faster_rcnn_x101_64x4d_fpn_2x_coco/faster_rcnn_x101_64x4d_fpn_2x_coco_20200512_161033-5961fa95.pth",
)



from icevision.visualize.utils import *
from icevision.visualize.draw_data import *
from icevision.visualize.show_data import *

from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.wandb:
    from icevision.visualize.wandb_img import *



__all__ = [
    "ImgSize",
    "open_img",
    "get_image_size",
    "get_img_size",
    "show_img",
    "plot_grid",
]

from icevision.imports import *
from PIL import ExifTags

ImgSize = namedtuple("ImgSize", "width,height")

# get exif tag
for _EXIF_ORIENTATION_TAG in ExifTags.TAGS.keys():
    if PIL.ExifTags.TAGS[_EXIF_ORIENTATION_TAG] == "Orientation":
        break

# from enum import Enum

# class PILMode(Enum):
#     blah

# FIXME
def open_img(fn, gray=False) -> PIL.Image.Image:
    "Open an image from disk `fn` as a PIL Image"
    color = "L" if gray else "RGB"
    image = PIL.Image.open(str(fn))
    image = PIL.ImageOps.exif_transpose(image)
    image = image.convert(color)
    return image


# TODO: Deprecated
def get_image_size(filepath: Union[str, Path]) -> Tuple[int, int]:
    """
    Returns image (width, height)
    """
    logger.warning("get_image_size is deprecated, use get_img_size instead")
    image_size = get_img_size(filepath=filepath)
    return image_size.width, image_size.height


def get_img_size(filepath: Union[str, Path]) -> ImgSize:
    """
    Returns image (width, height)
    """
    with PIL.Image.open(filepath) as image:
        image_size = image.size

    try:
        exif = image._getexif()
        if exif is not None and exif[_EXIF_ORIENTATION_TAG] in [6, 8]:
            image_size = image_size[::-1]
    except (AttributeError, KeyError, IndexError):
        # cases: image don't have getexif
        pass

    return ImgSize(*image_size)


def show_img(img, ax=None, show: bool = False, **kwargs):
    img = img.squeeze().copy()
    cmap = "gray" if len(img.shape) == 2 else None

    if ax is None:
        fig, ax = plt.subplots(**kwargs)

    ax.imshow(img, cmap=cmap)
    ax.set_axis_off()

    if show:
        plt.show()

    return ax


def plot_grid(
    fs: List[callable], ncols=1, figsize=None, show=False, axs_per_iter=1, **kwargs
):
    nrows = math.ceil(len(fs) * axs_per_iter / ncols)
    figsize = figsize or (12 * ncols, 12 * nrows)

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)
    axs = np.asarray(axs)

    if axs_per_iter == 1:
        axs = axs.flatten()
    elif axs_per_iter > 1:
        axs = axs.reshape(-1, axs_per_iter)
    else:
        raise ValueError("axs_per_iter has to be greater than 1")

    for f, ax in zip(fs, axs):
        f(ax=ax)

    plt.tight_layout()
    if show:
        plt.show()



__all__ = [
    "predict",
    "predict_from_dl",
    "convert_raw_prediction",
    "convert_raw_predictions",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.utils import _predict_from_dl
from icevision.models.mmdet.common.utils import *
from icevision.models.mmdet.common.mask.dataloaders import *
from icevision.models.mmdet.common.bbox.prediction import (
    _unpack_raw_bboxes,
)


@torch.no_grad()
def _predict_batch(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
):
    device = device or model_device(model)
    batch["img"] = [img.to(device) for img in batch["img"]]

    raw_preds = model(return_loss=False, rescale=False, **batch)
    return convert_raw_predictions(
        batch=batch,
        raw_preds=raw_preds,
        records=records,
        keep_images=keep_images,
        detection_threshold=detection_threshold,
    )


def predict(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    batch, records = build_infer_batch(dataset)

    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        keep_images=keep_images,
        device=device,
    )


def predict_from_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    keep_images: bool = False,
    **predict_kwargs,
):
    return _predict_from_dl(
        predict_fn=_predict_batch,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        keep_images=keep_images,
        **predict_kwargs,
    )


def convert_raw_predictions(
    batch,
    raw_preds,
    records: Sequence[BaseRecord],
    detection_threshold: float,
    keep_images: bool = False,
):

    # In inference, both "img" and "img_metas" are lists. Check out the `build_infer_batch()` definition
    # We need to convert that to a batch similar to train and valid batches
    if isinstance(batch["img"], list):
        batch = {
            "img": batch["img"][0],
            "img_metas": batch["img_metas"][0],
        }

    batch_list = [dict(zip(batch, t)) for t in zipsafe(*batch.values())]
    return [
        convert_raw_prediction(
            sample=sample,
            raw_pred=raw_pred,
            record=record,
            detection_threshold=detection_threshold,
            keep_image=keep_images,
        )
        for sample, raw_pred, record in zip(batch_list, raw_preds, records)
    ]


def convert_raw_prediction(
    sample,
    raw_pred: dict,
    record: BaseRecord,
    detection_threshold: float,
    keep_image: bool = False,
):
    # convert predictions
    raw_bboxes, raw_masks = raw_pred
    scores, labels, bboxes = _unpack_raw_bboxes(raw_bboxes)

    keep_mask = scores > detection_threshold
    keep_scores = scores[keep_mask]
    keep_labels = labels[keep_mask]
    keep_bboxes = [BBox.from_xyxy(*o) for o in bboxes[keep_mask]]
    keep_masks = MaskArray(np.vstack(raw_masks)[keep_mask])

    keep_labels = convert_background_from_last_to_zero(
        label_ids=keep_labels, class_map=record.detection.class_map
    )

    pred = BaseRecord(
        (
            ScoresRecordComponent(),
            ImageRecordComponent(),
            InstancesLabelsRecordComponent(),
            BBoxesRecordComponent(),
            MasksRecordComponent(),
        )
    )
    pred.detection.set_class_map(record.detection.class_map)
    pred.detection.set_scores(keep_scores)
    pred.detection.set_labels_by_id(keep_labels)
    pred.detection.set_bboxes(keep_bboxes)
    pred.detection.set_masks(keep_masks)
    pred.above_threshold = keep_mask

    if keep_image:
        image = mmdet_tensor_to_image(sample["img"])

        pred.set_img(image)
        record.set_img(image)

    return Prediction(pred=pred, ground_truth=record)



from icevision.models.ultralytics.yolov5.dataloaders import *
from icevision.models.ultralytics.yolov5.model import *
from icevision.models.ultralytics.yolov5.prediction import *
from icevision.models.ultralytics.yolov5.show_results import *
from icevision.models.ultralytics.yolov5.utils import *
from icevision.models.ultralytics.yolov5.backbones import *
from icevision.models.ultralytics.yolov5.show_batch import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    from icevision.models.ultralytics.yolov5 import fastai

if SoftDependencies.pytorch_lightning:
    from icevision.models.ultralytics.yolov5 import lightning



__all__ = ["ModelChoiceUI"]


from icevision import models
import ipywidgets as widgets
from IPython.display import display


class ModelChoiceUI:
    def __init__(self, task="object_detection"):
        self.task = task
        self.reset_lib_info()

        self.libraries_available = widgets.Dropdown(
            options=[],
            description="Libraries",
            disabled=False,
        )

        self.models_available = widgets.Dropdown(
            options=[""],
            description="Models",
            disabled=False,
        )

        self.backbones_available = widgets.Dropdown(
            options=[""],
            description="Backbones",
            disabled=False,
        )

    # lib_info = {
    #     "lib_type": None,
    #     "model_type": None,
    #     "backbone_type": None,
    #     "backbone": None,
    #     "model_list": [],
    #     "backbone_list": [],
    # }

    def reset_lib_info(self):
        self.lib_type = None
        self.model_type = None
        self.backbone_type = None
        self.model_list = []
        self.backbone_list = []
        self.backbone = None

    def get_model_info(self):
        return self.model_type, self.backbone

    # Creating dropdown widgets
    def populate_libraries(self):
        if self.task == "object_detection":
            libraries_list = ["", "MMDetection", "Ross Wightman", "Torchvision"]
        elif self.task == "mask":
            libraries_list = ["", "MMDetection", "Torchvision"]
        elif self.task == "keypoints":
            libraries_list = ["", "Torchvision"]

        self.libraries_available.options = libraries_list

    def od_library_change(self, change):
        lib_name = change.new

        if self.task == "object_detection":
            if lib_name == "Torchvision":
                lib_type = models.torchvision
                model_list = ["faster_rcnn", "retinanet"]

            if lib_name == "MMDetection":
                lib_type = models.mmdet.models
                model_list = ["retinanet", "faster_rcnn", "fcos", "sparse_rcnn"]

            if lib_name == "Ross Wightman":
                lib_type = models.ross
                model_list = ["efficientdet"]

        elif self.task == "mask":
            if lib_name == "Torchvision":
                lib_type = models.torchvision
                model_list = ["mask_rcnn"]

            if lib_name == "MMDetection":
                lib_type = models.mmdet.models
                model_list = ["mask_rcnn"]

        elif self.task == "keypoints":
            if lib_name == "Torchvision":
                lib_type = models.torchvision
                model_list = ["keypoint_rcnn"]

        self.lib_type = lib_type
        self.model_list = model_list

        self.models_available.options = model_list

    def od_model_change(self, change):
        model_name = change.new
        lib_type = self.lib_type
        model_type = getattr(lib_type, model_name)
        backbone_type = getattr(model_type, "backbones")
        backbone_list = [item for item in dir(model_type.backbones) if "__" not in item]

        self.model_type = model_type
        self.backbone_type = backbone_type
        self.backbone_list = backbone_list
        self.backbones_available.options = backbone_list

    def od_backbone_change(self, change):
        backbone_name = change.new
        model_type = self.model_type
        backbone_type = self.backbone_type

        backbone = getattr(backbone_type, backbone_name)

        self.backbone = backbone

    def display(self):
        self.populate_libraries()
        # Observe dropdown widget changes
        self.libraries_available.observe(self.od_library_change, names="value")
        self.models_available.observe(self.od_model_change, names="value")
        self.backbones_available.observe(self.od_backbone_change, names="value")

        # display dropdown widgets
        display(self.libraries_available)
        display(self.models_available)
        display(self.backbones_available)



__all__ = ["MaskMMDetectionCallback"]

from icevision.models.mmdet.fastai.callbacks import MMDetectionCallback
from icevision.models.mmdet.common.mask.prediction import convert_raw_predictions


class MaskMMDetectionCallback(MMDetectionCallback):
    def convert_raw_predictions(self, batch, raw_preds, records):
        return convert_raw_predictions(
            batch=batch, raw_preds=raw_preds, records=records, detection_threshold=0.0
        )



__all__ = ["model"]

from icevision.imports import *
from icevision.utils import *

import yaml
import yolov5
from yolov5.models.yolo import Model
from yolov5.utils.google_utils import attempt_download
from yolov5.utils.torch_utils import intersect_dicts
from yolov5.utils.general import check_img_size
from icevision.models.ultralytics.yolov5.utils import *
from icevision.models.ultralytics.yolov5.backbones import *

yolo_dir = get_root_dir() / "yolo"
yolo_dir.mkdir(exist_ok=True)


def model(
    backbone: YoloV5BackboneConfig,
    num_classes: int,
    img_size: int,  # must be multiple of 32
    device: Optional[torch.device] = None,
) -> nn.Module:
    model_name = backbone.model_name
    pretrained = backbone.pretrained

    # this is to remove background from ClassMap as discussed
    # here: https://github.com/ultralytics/yolov5/issues/2950
    # and here: https://discord.com/channels/735877944085446747/782062040168267777/836692604224536646
    # so we should pass `num_classes=parser.class_map.num_classes`
    num_classes -= 1

    device = (
        torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if device is None
        else device
    )

    if model_name in ["yolov5s", "yolov5m", "yolov5l", "yolov5x"]:
        cfg_filepath = Path(yolov5.__file__).parent / f"models/{model_name}.yaml"
    else:
        cfg_filepath = Path(yolov5.__file__).parent / f"models/hub/{model_name}.yaml"

    if pretrained:
        weights_path = yolo_dir / f"{model_name}.pt"

        with open(Path(yolov5.__file__).parent / "data/hyp.finetune.yaml") as f:
            hyp = yaml.load(f, Loader=yaml.SafeLoader)

        attempt_download(weights_path)  # download if not found locally
        sys.path.insert(0, str(Path(yolov5.__file__).parent))
        ckpt = torch.load(weights_path, map_location=device)  # load checkpoint
        sys.path.remove(str(Path(yolov5.__file__).parent))
        if hyp.get("anchors"):
            ckpt["model"].yaml["anchors"] = round(hyp["anchors"])  # force autoanchor
        model = Model(cfg_filepath or ckpt["model"].yaml, ch=3, nc=num_classes).to(
            device
        )  # create
        exclude = []  # exclude keys
        state_dict = ckpt["model"].float().state_dict()  # to FP32
        state_dict = intersect_dicts(
            state_dict, model.state_dict(), exclude=exclude
        )  # intersect
        model.load_state_dict(state_dict, strict=False)  # load
    else:
        with open(Path(yolov5.__file__).parent / "data/hyp.scratch.yaml") as f:
            hyp = yaml.load(f, Loader=yaml.SafeLoader)

        model = Model(
            cfg_filepath, ch=3, nc=num_classes, anchors=hyp.get("anchors")
        ).to(
            device
        )  # create

    gs = int(model.stride.max())  # grid size (max stride)
    nl = model.model[-1].nl  # number of detection layers (used for scaling hyp['obj'])
    imgsz = check_img_size(img_size, gs)  # verify imgsz are gs-multiples

    hyp["box"] *= 3.0 / nl  # scale to layers
    hyp["cls"] *= num_classes / 80.0 * 3.0 / nl  # scale to classes and layers
    hyp["obj"] *= (imgsz / 640) ** 2 * 3.0 / nl  # scale to image size and layers
    model.nc = num_classes  # attach number of classes to model
    model.hyp = hyp  # attach hyperparameters to model
    model.gr = 1.0  # iou loss ratio (obj_loss = 1.0 or iou)

    def param_groups_fn(model: nn.Module) -> List[List[nn.Parameter]]:
        spp_index = [
            i + 2
            for i, layer in enumerate(model.model.children())
            if layer._get_name() == "SPP"
        ][0]
        backbone = list(model.model.children())[:spp_index]
        neck = list(model.model.children())[spp_index:-1]
        head = list(model.model.children())[-1]

        layers = [nn.Sequential(*backbone), nn.Sequential(*neck), nn.Sequential(head)]

        param_groups = [list(group.parameters()) for group in layers]
        check_all_model_params_in_groups2(model.model, param_groups)

        return param_groups

    model.param_groups = MethodType(param_groups_fn, model)

    return model



from icevision.models.utils import *
from icevision.models.interpretation import *

# backwards compatibility
from icevision.models.torchvision import (
    faster_rcnn,
    mask_rcnn,
    retinanet,
    keypoint_rcnn,
)
from icevision.models import torchvision

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.effdet:
    # backwards compatibility
    from icevision.models.ross import efficientdet
    from icevision.models import ross

if SoftDependencies.mmdet:
    from icevision.models import mmdet

if SoftDependencies.yolov5:
    # HACK: yolov5 changes matplotlib backend here: https://github.com/ultralytics/yolov5/blob/77415a42e5975ea356393c9f1d5cff0ae8acae2c/utils/plots.py#L26
    import matplotlib

    backend = matplotlib.get_backend()
    from icevision.models import ultralytics

    matplotlib.use(backend)
    matplotlib.rcdefaults()



__all__ = [
    "resnet50_caffe_fpn_gn_head",
    "resnet50_caffe_fpn_gn_head_center_normbbox_centeronreg_giou",
    "resnet50_caffe_fpn_gn_head_dcn_1x_center_normbbox_centeronreg_giou",
    "resnet101_caffe_fpn_gn_head_1x_coco",
    "resnet50_caffe_fpn_gn_head_mstrain_640_800_2x",
    "resnet101_caffe_fpn_gn_head_mstrain_640_800_2x",
    "resnext101_64x4d_fpn_gn_head_mstrain_640_800_2x",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MMDetFCOSBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="fcos", **kwargs)


base_config_path = mmdet_configs_path / "fcos"
base_weights_url = (
    "https://openmmlab.oss-cn-hangzhou.aliyuncs.com/mmdetection/v2.0/fcos"
)

resnet50_caffe_fpn_gn_head = MMDetFCOSBackboneConfig(
    config_path=base_config_path / "fcos_r50_caffe_fpn_gn-head_1x_coco.py",
    weights_url=f"{base_weights_url}/fcos_r50_caffe_fpn_gn-head_1x_coco/fcos_r50_caffe_fpn_gn-head_1x_coco-821213aa.pth",
)

resnet50_caffe_fpn_gn_head_center_normbbox_centeronreg_giou = MMDetFCOSBackboneConfig(
    config_path=base_config_path
    / "fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_1x_coco.py",
    weights_url=f"{base_weights_url}/fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_1x_coco/fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_1x_coco-0a0d75a8.pth",
)

resnet50_caffe_fpn_gn_head_dcn_1x_center_normbbox_centeronreg_giou = MMDetFCOSBackboneConfig(
    config_path=base_config_path
    / "fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_dcn_1x_coco.py",
    weights_url=f"{base_weights_url}/fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_dcn_1x_coco/fcos_center-normbbox-centeronreg-giou_r50_caffe_fpn_gn-head_dcn_1x_coco-ae4d8b3d.pth",
)

resnet101_caffe_fpn_gn_head_1x_coco = MMDetFCOSBackboneConfig(
    config_path=base_config_path / "fcos_r101_caffe_fpn_gn-head_1x_coco.py",
    weights_url=f"{base_weights_url}/fcos_r101_caffe_fpn_gn-head_1x_coco/fcos_r101_caffe_fpn_gn-head_1x_coco-0e37b982.pth",
)

resnet50_caffe_fpn_gn_head_mstrain_640_800_2x = MMDetFCOSBackboneConfig(
    config_path=base_config_path
    / "fcos_r50_caffe_fpn_gn-head_mstrain_640-800_2x_coco.py",
    weights_url=f"{base_weights_url}/fcos_r50_caffe_fpn_gn-head_mstrain_640-800_2x_coco/fcos_r50_caffe_fpn_gn-head_mstrain_640-800_2x_coco-d92ceeea.pth",
)

resnet101_caffe_fpn_gn_head_mstrain_640_800_2x = MMDetFCOSBackboneConfig(
    config_path=base_config_path
    / "fcos_r101_caffe_fpn_gn-head_mstrain_640-800_2x_coco.py",
    weights_url=f"{base_weights_url}/fcos_r101_caffe_fpn_gn-head_mstrain_640-800_2x_coco/fcos_r101_caffe_fpn_gn-head_mstrain_640-800_2x_coco-511424d6.pth",
)

resnext101_64x4d_fpn_gn_head_mstrain_640_800_2x = MMDetFCOSBackboneConfig(
    config_path=base_config_path
    / "fcos_x101_64x4d_fpn_gn-head_mstrain_640-800_2x_coco.py",
    weights_url=f"{base_weights_url}/fcos_x101_64x4d_fpn_gn-head_mstrain_640-800_2x_coco/fcos_x101_64x4d_fpn_gn-head_mstrain_640-800_2x_coco-ede514a8.pth",
)



__all__ = ["MMDetectionCallback"]

from icevision.imports import *
from icevision.engines.fastai import *
from icevision.models.mmdet.utils import *

from icevision.models.mmdet.common.bbox.prediction import convert_raw_predictions


class _ModelWrap(nn.Module):
    def __init__(self, model: nn.Module):
        super().__init__()
        self.model = model

    def forward(self, xb):
        return self.model.train_step(data=xb, optimizer=None)

    def forward_test(self, xb):
        imgs = xb[0]["img"]
        img_metas = xb[0]["img_metas"]
        return self.model.forward_test(imgs=[imgs], img_metas=[img_metas])


class MMDetectionCallback(fastai.Callback):
    def after_create(self):
        self.learn.model = _ModelWrap(self.model)
        self.model.param_groups = self.model.model.param_groups

    def before_batch(self):
        self.learn.records = self.yb[0]
        self.learn.yb = self.xb

    @abstractmethod
    def convert_raw_predictions(self, batch, raw_preds, records):
        """Convert raw predictions from the model to library standard."""

    def after_loss(self):
        if not self.training:
            self.model.eval()
            with torch.no_grad():
                self.learn.raw_preds = self.model.forward_test(self.xb)
            self.model.train()

            # TODO: implement abstract function
            self.learn.converted_preds = self.convert_raw_predictions(
                batch=self.xb[0], raw_preds=self.raw_preds, records=self.records
            )



__all__ = [
    "predict",
    "predict_from_dl",
    "convert_raw_prediction",
    "convert_raw_predictions",
    "end2end_detect",
]

from icevision.models.torchvision.faster_rcnn.prediction import *



__all__ = ["download_mmdet_configs"]

from icevision.imports import *
from icevision.utils import *

VERSION = "v2.10.0"
BASE_URL = "https://codeload.github.com/airctic/mmdetection_configs/zip/refs/tags"


def download_mmdet_configs() -> Path:
    save_dir = get_root_dir() / f"mmdetection_configs"

    mmdet_config_path = save_dir / f"mmdetection_configs-{VERSION[1:]}/configs"
    download_path = save_dir / f"{VERSION}.zip"

    if mmdet_config_path.exists():
        logger.info(
            f"The mmdet config folder already exists. No need to downloaded it. Path : {mmdet_config_path}"
        )
    elif download_path.exists():
        # The zip file was downloaded by not extracted yet
        # Extract zip file
        logger.info(f"Extracting the {VERSION}.zip file.")
        save_dir = Path(download_path).parent
        shutil.unpack_archive(filename=str(download_path), extract_dir=str(save_dir))
    else:
        save_dir.mkdir(parents=True, exist_ok=True)

        download_path = save_dir / f"{VERSION}.zip"
        if not download_path.exists():
            logger.info("Downloading mmdet configs")
            download_and_extract(f"{BASE_URL}/{VERSION}", download_path)

    return mmdet_config_path



__all__ = [
    "predict",
    "predict_from_dl",
    "convert_raw_prediction",
    "convert_raw_predictions",
]

from itertools import chain
from icevision.imports import *
from icevision.core import *
from icevision.utils import *
from icevision.data import *
from icevision.models.utils import _predict_from_dl
from icevision.models.torchvision.keypoint_rcnn.dataloaders import *
from icevision.models.torchvision.faster_rcnn.prediction import (
    convert_raw_prediction as faster_convert_raw_prediction,
)


@torch.no_grad()
def _predict_batch(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
):
    model.eval()
    device = device or model_device(model)
    batch = [o.to(device) for o in batch]

    raw_preds = model(*batch)
    return convert_raw_predictions(
        batch=batch,
        raw_preds=raw_preds,
        records=records,
        detection_threshold=detection_threshold,
        keep_images=keep_images,
    )


def predict(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    batch, records = build_infer_batch(dataset)
    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        keep_images=keep_images,
        device=device,
    )


def predict_from_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    keep_images: bool = False,
    **predict_kwargs,
):
    return _predict_from_dl(
        predict_fn=_predict_batch,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        keep_images=keep_images,
        **predict_kwargs,
    )


def convert_raw_predictions(
    batch,
    raw_preds,
    records,
    detection_threshold: float,
    keep_images: bool = False,
):
    return [
        convert_raw_prediction(
            sample=sample,
            raw_pred=raw_pred,
            record=record,
            detection_threshold=detection_threshold,
            keep_image=keep_images,
        )
        for sample, raw_pred, record in zip(zip(*batch), raw_preds, records)
    ]


def convert_raw_prediction(
    sample,
    raw_pred: dict,
    record: BaseRecord,
    detection_threshold: float,
    keep_image: bool = False,
):
    pred = faster_convert_raw_prediction(
        sample=sample,
        raw_pred=raw_pred,
        record=record,
        detection_threshold=detection_threshold,
        keep_image=keep_image,
    )

    above_threshold = pred.detection.above_threshold
    kps = raw_pred["keypoints"][above_threshold]
    keypoints = []
    for k in kps:
        k = k.cpu().numpy()
        k = list(chain.from_iterable(k))
        # `if sum(k) > 0` prevents empty `KeyPoints` objects to be instantiated.
        # E.g. `k = [0, 0, 0, 0, 0, 0]` is a flattened list of 2 points `(0, 0, 0)` and `(0, 0, 0)`. We don't want a `KeyPoints` object to be created on top of this list.
        if sum(k) > 0:
            keypoints.append(KeyPoints.from_xyv(k, None))

    pred.pred.add_component(KeyPointsRecordComponent())
    pred.pred.detection.add_keypoints(keypoints)
    pred.pred.detection.keypoints_scores = (
        raw_pred["keypoints_scores"][above_threshold].detach().cpu().numpy()
    )

    return pred



__all__ = ["ModelAdapter"]

from icevision.imports import *
from icevision.metrics import *
from icevision.engines.lightning.lightning_model_adapter import LightningModelAdapter
from icevision.models.ultralytics import yolov5
from yolov5.utils.loss import ComputeLoss


class ModelAdapter(LightningModelAdapter, ABC):
    """Lightning module specialized for EfficientDet, with metrics support.

    The methods `forward`, `training_step`, `validation_step`, `validation_epoch_end`
    are already overriden.

    # Arguments
        model: The pytorch model to use.
        metrics: `Sequence` of metrics to use.

    # Returns
        A `LightningModule`.
    """

    def __init__(self, model: nn.Module, metrics: List[Metric] = None):
        super().__init__(metrics=metrics)
        self.model = model
        self.compute_loss = ComputeLoss(model)

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        (xb, yb), _ = batch
        preds = self(xb)

        loss = self.compute_loss(preds, yb)[0]

        self.log("train_loss", loss)

        return loss

    def validation_step(self, batch, batch_idx):
        (xb, yb), records = batch

        with torch.no_grad():
            inference_out, training_out = self(xb)
            preds = yolov5.convert_raw_predictions(
                batch=xb,
                raw_preds=inference_out,
                records=records,
                detection_threshold=0.001,
                nms_iou_threshold=0.6,
            )
            loss = self.compute_loss(training_out, yb)[0]

        self.accumulate_metrics(preds)

        self.log("val_loss", loss)

    def validation_epoch_end(self, outs):
        self.finalize_metrics()



__all__ = [
    "tf_lite0",
    "tf_lite1",
    "tf_lite2",
    "tf_lite3",
    "tf_d0",
    "tf_d1",
    "tf_d2",
    "tf_d3",
    "tf_d4",
    "tf_d5",
    "tf_d6",
    "tf_d7",
    "tf_d7x",
    "d0",
    "d1",
    "d2",
    "d3",
    "d4",
    "d5",
    "d6",
    "d7",
    "d7x",
    "tf_d0_ap",
    "tf_d1_ap",
    "tf_d2_ap",
    "tf_d3_ap",
    "tf_d4_ap",
    "tf_d5_ap",
]

from icevision.models.ross.efficientdet.utils import *


tf_lite0 = EfficientDetBackboneConfig(model_name="tf_efficientdet_lite0")
tf_lite1 = EfficientDetBackboneConfig(model_name="tf_efficientdet_lite1")
tf_lite2 = EfficientDetBackboneConfig(model_name="tf_efficientdet_lite2")
tf_lite3 = EfficientDetBackboneConfig(model_name="tf_efficientdet_lite3")

tf_d0 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d0")
tf_d1 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d1")
tf_d2 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d2")
tf_d3 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d3")
tf_d4 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d4")
tf_d5 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d5")
tf_d6 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d6")
tf_d7 = EfficientDetBackboneConfig(model_name="tf_efficientdet_d7")
tf_d7x = EfficientDetBackboneConfig(model_name="tf_efficientdet_d7x")

d0 = EfficientDetBackboneConfig(model_name="efficientdet_d0")
d1 = EfficientDetBackboneConfig(model_name="efficientdet_d1")
d2 = EfficientDetBackboneConfig(model_name="efficientdet_d2")
d3 = EfficientDetBackboneConfig(model_name="efficientdet_d3")
d4 = EfficientDetBackboneConfig(model_name="efficientdet_d4")
d5 = EfficientDetBackboneConfig(model_name="efficientdet_d5")
d6 = EfficientDetBackboneConfig(model_name="efficientdet_d6")
d7 = EfficientDetBackboneConfig(model_name="efficientdet_d7")
d7x = EfficientDetBackboneConfig(model_name="efficientdet_d7x")

tf_d0_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d0_ap")
tf_d1_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d1_ap")
tf_d2_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d2_ap")
tf_d3_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d3_ap")
tf_d4_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d4_ap")
tf_d5_ap = EfficientDetBackboneConfig(model_name="tf_efficientdet_d5_ap")



from icevision.imports import *
from icevision import BBox, BaseRecord


def get_best_score_item(prediction_items: Collection[Dict]):
    # fill with dummy if list of prediction_items is empty
    dummy = dict(
        predicted_bbox=BBox.from_xyxy(0, 0, 0, 0),
        score=1.0,
        iou_score=1.0,
        predicted_label_id=0,
    )
    best_item = max(prediction_items, key=lambda x: x["score"], default=dummy)
    return best_item


def pairwise_iou_record_record(target: BaseRecord, prediction: BaseRecord):
    """
    Calculates pairwise iou on prediction and target BaseRecord. Uses torchvision implementation of `box_iou`.
    """
    stacked_preds = [bbox.to_tensor() for bbox in prediction.detection.bboxes]
    stacked_preds = torch.stack(stacked_preds) if stacked_preds else torch.empty(0, 4)

    stacked_targets = [bbox.to_tensor() for bbox in target.detection.bboxes]
    stacked_targets = (
        torch.stack(stacked_targets) if stacked_targets else torch.empty(0, 4)
    )
    return torchvision.ops.box_iou(stacked_preds, stacked_targets)


def match_records(
    target: BaseRecord, prediction: BaseRecord, iou_threshold: float = 0.5
) -> Collection:
    """
    matches bboxes, labels from targets with their predictions by iou threshold
    """
    # here we get a tensor of indices that match iou criteria (order is (pred_id, target_id))
    iou_table = pairwise_iou_record_record(target=target, prediction=prediction)
    pairs_indices = torch.nonzero(iou_table > iou_threshold)

    # creating a list of [target, matching_predictions]
    target_list = [
        [dict(target_bbox=bbox, target_label=label, target_label_id=label_id), []]
        for bbox, label, label_id in zip(
            target.detection.bboxes, target.detection.labels, target.detection.label_ids
        )
    ]
    prediction_list = [
        dict(
            predicted_bbox=bbox,
            predicted_label=label,
            predicted_label_id=label_id,
            score=score,
        )
        for bbox, label, label_id, score in zip(
            prediction.detection.bboxes,
            prediction.detection.labels,
            prediction.detection.label_ids,
            prediction.detection.scores,
        )
    ]

    # appending matches to targets
    for pred_id, target_id in pairs_indices:
        single_prediction = deepcopy(prediction_list[pred_id])
        # python value casting needs rounding cause otherwise there are 0.69999991 values
        iou_score = round(iou_table[pred_id, target_id].item(), 4)
        single_prediction["iou_score"] = iou_score
        # seems like a magic number, but we want to append to the list of target's matching_predictions
        target_list[target_id][1].append(single_prediction)

    return target_list



__all__ = ["aggregate_records_objects"]

from icevision.imports import *
from .record import BaseRecord


def aggregate_records_objects(
    records: List[BaseRecord],
) -> Dict[str, List[Dict[str, Any]]]:
    objects = defaultdict(lambda: defaultdict(list))
    for record in records:
        for composite_name, composite_dict in record.aggregate_objects().items():
            for component_name, component_value in composite_dict.items():
                objects[composite_name][component_name].extend(component_value)
    return dict(objects)



__all__ = ["ModelAdapter"]

from icevision.models.mmdet.lightning.model_adapter import MMDetModelAdapter
from icevision.models.mmdet.common.bbox import convert_raw_predictions


class ModelAdapter(MMDetModelAdapter):
    def convert_raw_predictions(self, batch, raw_preds, records):
        return convert_raw_predictions(
            batch=batch, raw_preds=raw_preds, records=records, detection_threshold=0.0
        )



__all__ = [
    "filter_params",
    "unfreeze",
    "freeze",
    "transform_dl",
    "apply_batch_tfms",
    "_predict_from_dl",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.parsers import *

BN_TYPES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def filter_params(
    module: nn.Module, bn: bool = True, only_trainable=False
) -> Generator:
    """Yields the trainable parameters of a given module.

    Args:
        module: A given module
        bn: If False, don't return batch norm layers

    Returns:
        Generator
    """
    children = list(module.children())
    if not children:
        if not isinstance(module, BN_TYPES) or bn:
            for param in module.parameters():
                if not only_trainable or param.requires_grad:
                    yield param
    else:
        for child in children:
            for param in filter_params(
                module=child, bn=bn, only_trainable=only_trainable
            ):
                yield param


def unfreeze(params):
    for p in params:
        p.requires_grad = True


def freeze(params):
    for p in params:
        p.requires_grad = False


def transform_dl(
    dataset,
    build_batch,
    batch_tfms=None,
    build_batch_kwargs: dict = {},
    **dataloader_kwargs,
):
    """Creates collate_fn from build_batch (collate function) by decorating it with apply_batch_tfms and unload_records"""
    collate_fn = apply_batch_tfms(
        build_batch, batch_tfms=batch_tfms, **build_batch_kwargs
    )
    collate_fn = unload_records(collate_fn)
    return DataLoader(dataset=dataset, collate_fn=collate_fn, **dataloader_kwargs)


def apply_batch_tfms(build_batch, batch_tfms=None, **build_batch_kwargs):
    """This decorator function applies batch_tfms to records before passing them to build_batch"""

    def inner(records):
        if batch_tfms is not None:
            records = batch_tfms(records)
        return build_batch(records, **build_batch_kwargs)

    return inner


def unload_records(build_batch):
    """This decorator function unloads records to not carry them around after batch creation"""

    def inner(records):
        tupled_output, records = build_batch(records)
        for record in records:
            record.unload()
        return tupled_output, records

    return inner


@torch.no_grad()
def _predict_from_dl(
    predict_fn,
    model: nn.Module,
    infer_dl: DataLoader,
    keep_images: bool = False,
    show_pbar: bool = True,
    **predict_kwargs,
) -> List[Prediction]:
    all_preds = []
    for batch, records in pbar(infer_dl, show=show_pbar):
        preds = predict_fn(
            model=model,
            batch=batch,
            records=records,
            keep_images=keep_images,
            **predict_kwargs,
        )
        all_preds.extend(preds)

    return all_preds



__all__ = ["predict", "predict_from_dl", "convert_raw_predictions", "end2end_detect"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.utils import _predict_from_dl
from icevision.models.ultralytics.yolov5.dataloaders import *
from yolov5.utils.general import non_max_suppression
from icevision.models.inference import *


@torch.no_grad()
def _predict_batch(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.25,
    nms_iou_threshold: float = 0.45,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    # device issue addressed on discord: https://discord.com/channels/735877944085446747/770279401791160400/832361687855923250
    if device is not None:
        raise ValueError(
            "For YOLOv5 device can only be specified during model creation, "
            "for more info take a look at the discussion here: "
            "https://discord.com/channels/735877944085446747/770279401791160400/832361687855923250"
        )
    grid = model.model[-1].grid[-1]
    # if `grid.numel() == 1` it means the grid isn't initialized yet and we can't
    # trust it's device (will always be CPU)
    device = grid.device if grid.numel() > 1 else model_device(model)

    batch = batch[0].to(device)
    model = model.eval().to(device)

    raw_preds = model(batch)[0]
    return convert_raw_predictions(
        batch=batch,
        raw_preds=raw_preds,
        records=records,
        detection_threshold=detection_threshold,
        nms_iou_threshold=nms_iou_threshold,
        keep_images=keep_images,
    )


def predict(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.25,
    nms_iou_threshold: float = 0.45,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    batch, records = build_infer_batch(dataset)
    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        nms_iou_threshold=nms_iou_threshold,
        keep_images=keep_images,
        device=device,
    )


def predict_from_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    keep_images: bool = False,
    **predict_kwargs,
):
    return _predict_from_dl(
        predict_fn=_predict_batch,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        keep_images=keep_images,
        **predict_kwargs,
    )


def convert_raw_predictions(
    batch,
    raw_preds: torch.Tensor,
    records: Sequence[BaseRecord],
    detection_threshold: float,
    nms_iou_threshold: float,
    keep_images: bool = False,
) -> List[Prediction]:
    dets = non_max_suppression(
        raw_preds, conf_thres=detection_threshold, iou_thres=nms_iou_threshold
    )
    dets = [d.detach().cpu().numpy() for d in dets]
    preds = []
    for det, record, tensor_image in zip(dets, records, batch):

        pred = BaseRecord(
            (
                ScoresRecordComponent(),
                ImageRecordComponent(),
                InstancesLabelsRecordComponent(),
                BBoxesRecordComponent(),
            )
        )

        pred.detection.set_class_map(record.detection.class_map)
        pred.detection.set_labels_by_id(det[:, 5].astype(int) + 1)
        pred.detection.set_bboxes([BBox.from_xyxy(*xyxy) for xyxy in det[:, :4]])
        pred.detection.set_scores(det[:, 4])

        if keep_images:
            record.set_img(tensor_to_image(tensor_image))

        preds.append(Prediction(pred=pred, ground_truth=record))

    return preds


end2end_detect = partial(_end2end_detect, predict_fn=predict)



__all__ = ["SimpleConfusionMatrix"]

from icevision.data.prediction import Prediction
from icevision.metrics.metric import Metric
from icevision.imports import *
from icevision.metrics.confusion_matrix.confusion_matrix_utils import *
import PIL


class MatchingPolicy(Enum):
    BEST_SCORE = 1
    BEST_IOU = 2


class SimpleConfusionMatrix(Metric):
    def __init__(
        self,
        iou_threshold: float = 0.5,
        policy: MatchingPolicy = MatchingPolicy.BEST_SCORE,
        print_summary: bool = False,
    ):
        super(SimpleConfusionMatrix, self).__init__()
        self.print_summary = print_summary
        self.target_labels = []
        self.predicted_labels = []
        self._iou_threshold = iou_threshold
        self._policy = policy
        self.class_map = None
        self.confusion_matrix: sklearn.metrics.confusion_matrix = None

    def _reset(self):
        self.target_labels = []
        self.predicted_labels = []

    def accumulate(self, preds: Collection[Prediction]):
        for pred in preds:
            target_record = pred.ground_truth
            prediction_record = pred.pred
            self.class_map = target_record.detection.class_map
            # skip if empty ground_truths
            # if not target_record.detection.bboxes:
            #     continue
            # create matches based on iou
            matches = match_records(
                target=target_record,
                prediction=prediction_record,
                iou_threshold=self._iou_threshold,
            )

            target_labels, predicted_labels = [], []
            # iterate over multiple targets and preds in a record
            for target_item, prediction_items in matches:
                if self._policy == MatchingPolicy.BEST_SCORE:
                    predicted_item = get_best_score_item(
                        prediction_items=prediction_items,
                    )
                elif self._policy == MatchingPolicy.BEST_IOU:
                    raise NotImplementedError
                else:
                    raise RuntimeError(f"policy must be one of {list(MatchingPolicy)}")

                # using label_id instead of named label to save memory
                target_label = target_item["target_label_id"]
                predicted_label = predicted_item["predicted_label_id"]
                target_labels.append(target_label)
                predicted_labels.append(predicted_label)

            # We need to store the entire list of gts/preds to support various CM logging methods
            assert len(predicted_labels) == len(target_labels)
            self.target_labels.extend(target_labels)
            self.predicted_labels.extend(predicted_labels)

    def finalize(self):
        """Convert preds to numpy arrays and calculate the CM"""
        assert len(self.target_labels) == len(self.predicted_labels)
        label_ids = list(self.class_map._class2id.values())
        self.confusion_matrix = sklearn.metrics.confusion_matrix(
            y_true=self.target_labels,
            y_pred=self.predicted_labels,
            labels=label_ids,
        )
        if self.print_summary:
            print(self.confusion_matrix)
        self._reset()
        return {"dummy_value_for_fastai": -1}

    def plot(
        self,
        normalize: Optional[str] = None,
        xticks_rotation="vertical",
        values_format: str = None,
        cmap: str = "PuBu",
        figsize: int = 11,
        **display_args,
    ):
        """
        A handle to plot the matrix in a jupyter notebook, potentially this could also be passed to save_fig
        """
        if normalize not in ["true", "pred", "all", None]:
            raise ValueError(
                "normalize must be one of {'true', 'pred', " "'all', None}"
            )
        # properly display ints and floats
        if values_format is not None:
            values_format = ".2f" if normalize else "d"

        cm = self._maybe_normalize(self.confusion_matrix, normalize)
        labels_named = self.class_map._id2class
        cm_display = sklearn.metrics.ConfusionMatrixDisplay(
            cm, display_labels=labels_named
        )
        figure = cm_display.plot(
            xticks_rotation=xticks_rotation,
            cmap=cmap,
            values_format=values_format,
            **display_args,
        ).figure_
        figure.set_size_inches(figsize, figsize)
        plt.close()
        return figure

    def _fig2img(self, fig):
        # TODO: start using fi2img from icevision utils
        """Converts matplotlib figure object to PIL Image for easier logging. Writing to buffer is necessary
        to avoid wandb cutting our labels off. Wandb autoconvert doesn't pass the `bbox_inches` parameter so we need
        to do this manually."""
        buf = io.BytesIO()
        fig.savefig(buf, bbox_inches="tight")
        buf.seek(0)
        return PIL.Image.open(buf)

    def _maybe_normalize(self, cm, normalize):
        """This method is copied from sklearn. Only used in plot_confusion_matrix but we want to be able
        to normalize upon plotting."""
        with np.errstate(all="ignore"):
            if normalize == "true":
                cm = cm / cm.sum(axis=1, keepdims=True)
            elif normalize == "pred":
                cm = cm / cm.sum(axis=0, keepdims=True)
            elif normalize == "all":
                cm = cm / cm.sum()
        cm = np.nan_to_num(cm)
        return cm

    def log(self, logger_object) -> None:
        # TODO: Disabled for now, need to design for metric logging for this to work + pl dependency
        # if isinstance(logger_object, pl_loggers.WandbLogger):
        #     fig = self.plot()
        #     image = self._fig2img(fig)
        #     logger_object.experiment.log({"Confusion Matrix": wandb.Image(image)})
        return



__all__ = [
    "ssd300",
    "ssd512",
]

from icevision.imports import *
from icevision.models.mmdet.utils import *


class MMDetSSDBackboneConfig(MMDetBackboneConfig):
    def __init__(self, **kwargs):
        super().__init__(model_name="ssd", **kwargs)


base_config_path = mmdet_configs_path / "ssd"
base_weights_url = "http://download.openmmlab.com/mmdetection/v2.0/ssd"

ssd300 = MMDetSSDBackboneConfig(
    config_path=base_config_path / "ssd300_coco.py",
    weights_url=f"{base_weights_url}/ssd300_coco/ssd300_coco_20200307-a92d2092.pth",
)

ssd512 = MMDetSSDBackboneConfig(
    config_path=base_config_path / "ssd512_coco.py",
    weights_url=f"{base_weights_url}/ssd512_coco/ssd512_coco_20200308-038c5591.pth",
)



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from icevision.imports import *
from icevision.core import *
from icevision.models.utils import *


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for training the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def _build_train_sample(
    record: RecordType,
) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
    assert len(record.detection.label_ids) == len(record.detection.bboxes)

    image = im2tensor(record.img)

    # If no labels and bboxes are present, use as negative samples
    if len(record.detection.label_ids) == 0:
        target = torch.zeros((0, 6))
    else:
        labels = tensor(record.detection.label_ids, dtype=torch.int64) - 1

        img_width, img_height = record.width, record.height
        xyxys = [
            bbox.relative_xcycwh(img_width, img_height)
            for bbox in record.detection.bboxes
        ]
        boxes = tensor(xyxys, dtype=torch.float32)

        target = torch.zeros((len(labels), 6))
        target[:, 1:] = torch.cat([labels.unsqueeze(1), boxes], 1)

    return image, target


def build_train_batch(
    records: Sequence[RecordType],
) -> Tuple[List[torch.Tensor], List[Dict[str, torch.Tensor]]]:
    """Builds a batch in the format required by the model when training.

    # Arguments
        records: A `Sequence` of records.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_train_batch(records)
    outs = model(*batch)
    ```
    """
    images, targets = [], []
    for i, record in enumerate(records):
        image, target = _build_train_sample(record)
        images.append(image)

        if target.numel() > 0:
            target[:, 0] = i

        targets.append(target)

    return (torch.stack(images, 0), torch.cat(targets, 0)), records


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for validating the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_valid_batch(
    records: List[RecordType],
) -> Tuple[List[torch.Tensor], Dict[str, torch.Tensor]]:
    """Builds a batch in the format required by the model when validating.

    # Arguments
        records: A `Sequence` of records.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_valid_batch(records)
    outs = model(*batch)
    ```
    """
    return build_train_batch(records=records)


def infer_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for inferring the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_infer_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_infer_batch(records: Sequence[RecordType]):
    """Builds a batch in the format required by the model when doing inference.

    # Arguments
        records: A `Sequence` of records.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_infer_batch(records)
    outs = model(*batch)
    ```
    """
    tensor_imgs = [im2tensor(record.img) for record in records]
    tensor_imgs = torch.stack(tensor_imgs)

    return (tensor_imgs,), records



__all__ = ["MaskRCNNCallback"]

from icevision.models.torchvision.fastai.callbacks import *
from icevision.models.torchvision.mask_rcnn.prediction import *


class MaskRCNNCallback(RCNNCallback):
    def convert_raw_predictions(self, batch, raw_preds):
        return convert_raw_predictions(
            batch=batch,
            raw_preds=raw_preds,
            records=self.learn.records,
            detection_threshold=0.0,
            mask_threshold=0.0,
        )



__all__ = ["Yolov5Callback"]

from icevision.engines.fastai import *
from icevision.models.ultralytics import yolov5


class Yolov5Callback(fastai.Callback):
    def before_batch(self):
        assert len(self.xb) == len(self.yb) == 1, "Only works for single input-output"
        x, y, records = self.xb[0][0], self.xb[0][1], self.yb
        self.learn.xb = [x]
        self.learn.yb = [y]
        self.learn.records = records[0]

    def after_pred(self):
        if not self.training:
            inference_out, training_out = self.pred[0], self.pred[1]
            self.learn.pred = training_out
            batch = (*self.learn.xb, *self.learn.yb)
            preds = yolov5.convert_raw_predictions(
                batch=batch,
                raw_preds=inference_out,
                records=self.learn.records,
                detection_threshold=0.001,
                nms_iou_threshold=0.6,
            )
            self.learn.converted_preds = preds



__all__ = [
    "small",
    "small_p6",
    "medium",
    "medium_p6",
    "large",
    "large_p6",
    "extra_large",
    "extra_large_p6",
]

from icevision.models.ultralytics.yolov5.utils import *


small = YoloV5BackboneConfig(model_name="yolov5s")
small_p6 = YoloV5BackboneConfig(model_name="yolov5s6")

medium = YoloV5BackboneConfig(model_name="yolov5m")
medium_p6 = YoloV5BackboneConfig(model_name="yolov5m6")

large = YoloV5BackboneConfig(model_name="yolov5l")
large_p6 = YoloV5BackboneConfig(model_name="yolov5l6")

extra_large = YoloV5BackboneConfig(model_name="yolov5x")
extra_large_p6 = YoloV5BackboneConfig(model_name="yolov5x6")



__all__ = ["show_batch"]

from icevision.utils import *
from icevision.visualize import *
from icevision.models.mmdet.common.utils import *


def show_batch(batch_and_records, ncols: int = 1, figsize=None, **show_samples_kwargs):
    """Show a single batch from a dataloader.
    # Arguments
        show_samples_kwargs: Check the parameters from `show_samples`
    """
    batch, records = batch_and_records

    # In inference, both "img" and "img_metas" are lists. Check out the `build_infer_batch()` definition
    # We need to convert that to a batch similar to train and valid batches
    if isinstance(batch["img"], list):
        batch = {
            "img": batch["img"][0],
            "img_metas": batch["img_metas"][0],
        }

    for tensor_image, record in zip(batch["img"][:], records):
        image = mmdet_tensor_to_image(tensor_image)
        record.set_img(image)

    return show_samples(records, ncols=ncols, figsize=figsize, **show_samples_kwargs)



__all__ = ["wandb_img_preds", "wandb_image"]


from typing import List

import wandb
from icevision import BaseRecord, BBox
from icevision.data.prediction import Prediction


def wandb_img_preds(
    preds: List[Prediction], add_ground_truth: bool = False
) -> List[wandb.Image]:
    return [wandb_image(pred, add_ground_truth=add_ground_truth) for pred in preds]


def bbox_wandb(bbox: BBox, label_id: int, label_name: str, score=None) -> dict:
    """Return a wandb compatible dictionary with bbox, label and score"""
    xmin, ymin, xmax, ymax = map(int, bbox.xyxy)

    box_data = {
        "position": {"minX": xmin, "maxX": xmax, "minY": ymin, "maxY": ymax},
        "class_id": int(label_id),
        "domain": "pixel",
    }

    if score:
        score = int(score * 100)
        box_caption = f"{label_name} ({score}%)"
        box_data["score"] = score
    else:
        box_caption = label_name

    box_data["box_caption"] = box_caption

    return box_data


def wandb_image(pred: Prediction, add_ground_truth: bool = False) -> wandb.Image:
    """Return a wandb image corresponding to the a prediction.

    Args:
        pred (Prediction): A prediction to log with WandB.
            Must have been created with keep_image = True.
        add_ground_truth (bool, optional): Add ground_truth information to the
            the WandB image. Defaults to False.

    Returns:
        wandb.Image: Specifying the image, but also the predictions and  possibly ground_truth.
    """
    # FIXME: if pred does not have an img, then we lose.
    # FIXME: Not handling masks

    # Check if "masks" key is the sample dictionnary
    # if "masks" in sample:     true_masks = sample["masks"]

    # Check if "masks" key is the pred dictionnary
    # if "masks" in pred: pred_masks = pred["masks"]

    class_id_to_label = {
        id: label for id, label in enumerate(pred.detection.class_map._id2class)
    }

    # Prediction
    box_data = list(
        map(
            bbox_wandb,
            pred.detection.bboxes,
            pred.detection.label_ids,
            pred.detection.labels,
            pred.detection.scores,
        )
    )

    boxes = {"predictions": {"box_data": box_data, "class_labels": class_id_to_label}}

    # Predicted Masks
    # Check if "masks" key is the pred dictionnary
    #     if "masks" in pred:
    #         mask_data = (pred_masks.data * pred["labels"][:, None, None]).max(0)
    #         masks = {
    #             "predictions": {"mask_data": mask_data, "class_labels": class_id_to_label}
    #         }
    #     else:
    #         masks = None
    masks = None

    # Ground Truth
    if add_ground_truth:
        box_data = list(
            map(
                bbox_wandb,
                pred.ground_truth.detection.bboxes,
                pred.ground_truth.detection.label_ids,
                pred.ground_truth.detection.labels,
            )
        )

        boxes["ground_truth"] = {
            "box_data": box_data,
            "class_labels": class_id_to_label,
        }

        # # Ground Truth Masks
        # Check if "masks" key is the sample dictionnary
        #         if "masks" in sample:
        #             labels_arr = np.array(sample["labels"])
        #             mask_data = (true_masks.data * labels_arr[:, None, None]).max(0)
        #             masks["ground_truth"] = {
        #                 "mask_data": mask_data,
        #                 "class_labels": class_id_to_label,
        #             }
    return wandb.Image(pred.img, boxes=boxes, masks=masks)



__all__ = [
    "predict",
    "predict_from_dl",
    "convert_raw_prediction",
    "convert_raw_predictions",
    "end2end_detect",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.utils import _predict_from_dl
from icevision.models.mmdet.common.utils import *
from icevision.models.mmdet.common.bbox.dataloaders import build_infer_batch
from icevision.models.mmdet.common.utils import convert_background_from_last_to_zero
from icevision.models.inference import *


@torch.no_grad()
def _predict_batch(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    records: Sequence[BaseRecord],
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
):
    device = device or model_device(model)
    batch["img"] = [img.to(device) for img in batch["img"]]

    raw_preds = model(return_loss=False, rescale=False, **batch)
    return convert_raw_predictions(
        batch=batch,
        raw_preds=raw_preds,
        records=records,
        keep_images=keep_images,
        detection_threshold=detection_threshold,
    )


def predict(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.5,
    keep_images: bool = False,
    device: Optional[torch.device] = None,
) -> List[Prediction]:
    batch, records = build_infer_batch(dataset)

    return _predict_batch(
        model=model,
        batch=batch,
        records=records,
        detection_threshold=detection_threshold,
        keep_images=keep_images,
        device=device,
    )


end2end_detect = partial(_end2end_detect, predict_fn=predict)


def predict_from_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    keep_images: bool = False,
    **predict_kwargs
):
    _predict_batch_fn = partial(_predict_batch, keep_images=keep_images)
    return _predict_from_dl(
        predict_fn=_predict_batch_fn,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        keep_images=keep_images,
        **predict_kwargs,
    )


def convert_raw_predictions(
    batch,
    raw_preds,
    records: Sequence[BaseRecord],
    detection_threshold: float,
    keep_images: bool = False,
):

    # In inference, both "img" and "img_metas" are lists. Check out the `build_infer_batch()` definition
    # We need to convert that to a batch similar to train and valid batches
    if isinstance(batch["img"], list):
        batch = {
            "img": batch["img"][0],
            "img_metas": batch["img_metas"][0],
        }

    batch_list = [dict(zip(batch, t)) for t in zipsafe(*batch.values())]
    return [
        convert_raw_prediction(
            sample=sample,
            raw_pred=raw_pred,
            record=record,
            detection_threshold=detection_threshold,
            keep_image=keep_images,
        )
        for sample, raw_pred, record in zip(batch_list, raw_preds, records)
    ]


def convert_raw_prediction(
    sample,
    raw_pred: dict,
    record: BaseRecord,
    detection_threshold: float,
    keep_image: bool = False,
):
    # convert predictions
    raw_bboxes = raw_pred
    scores, labels, bboxes = _unpack_raw_bboxes(raw_bboxes)

    keep_mask = scores > detection_threshold
    keep_scores = scores[keep_mask]
    keep_labels = labels[keep_mask]
    keep_bboxes = [BBox.from_xyxy(*o) for o in bboxes[keep_mask]]

    keep_labels = convert_background_from_last_to_zero(
        label_ids=keep_labels, class_map=record.detection.class_map
    )

    pred = BaseRecord(
        (
            ScoresRecordComponent(),
            ImageRecordComponent(),
            InstancesLabelsRecordComponent(),
            BBoxesRecordComponent(),
        )
    )
    pred.detection.set_class_map(record.detection.class_map)
    pred.detection.set_scores(keep_scores)
    pred.detection.set_labels_by_id(keep_labels)
    pred.detection.set_bboxes(keep_bboxes)
    pred.above_threshold = keep_mask

    if keep_image:
        image = mmdet_tensor_to_image(sample["img"])

        pred.set_img(image)
        record.set_img(image)

    return Prediction(pred=pred, ground_truth=record)


def _unpack_raw_bboxes(raw_bboxes):
    stack_raw_bboxes = np.vstack(raw_bboxes)

    scores = stack_raw_bboxes[:, -1]
    bboxes = stack_raw_bboxes[:, :-1]

    # each item in raw_pred is an array of predictions of it's `i` class
    labels = [np.full(o.shape[0], i, dtype=np.int32) for i, o in enumerate(raw_bboxes)]
    labels = np.concatenate(labels)

    return scores, labels, bboxes



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from mmdet.core import BitmapMasks
from icevision.core import *
from icevision.imports import *
from icevision.models.utils import *
from icevision.models.mmdet.common.bbox.dataloaders import (
    _img_tensor,
    _img_meta,
    _labels,
    _bboxes,
)


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def infer_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for inferring the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_infer_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_valid_batch(
    records: Sequence[RecordType],
) -> Tuple[dict, List[Dict[str, torch.Tensor]]]:
    return build_train_batch(records=records)


def build_train_batch(
    records: Sequence[RecordType],
) -> Tuple[dict, List[Dict[str, torch.Tensor]]]:
    images, labels, bboxes, masks, img_metas = [], [], [], [], []
    for record in records:
        images.append(_img_tensor(record))
        img_metas.append(_img_meta_mask(record))
        labels.append(_labels(record))
        bboxes.append(_bboxes(record))
        masks.append(_masks(record))

    data = {
        "img": torch.stack(images),
        "img_metas": img_metas,
        "gt_labels": labels,
        "gt_bboxes": bboxes,
        "gt_masks": masks,
    }

    return data, records


def build_infer_batch(records):
    imgs, img_metas = [], []
    for record in records:
        imgs.append(_img_tensor(record))
        img_metas.append(_img_meta_mask(record))

    data = {
        "img": [torch.stack(imgs)],
        "img_metas": [img_metas],
    }

    return data, records


def _img_meta_mask(record):
    img_meta = _img_meta(record)
    img_meta["ori_shape"] = img_meta["pad_shape"]
    return img_meta


def _masks(record):
    if len(record.detection.masks) == 0:
        raise RuntimeError("Negative samples still needs to be implemented")
    else:
        mask = record.detection.masks.data
        _, h, w = mask.shape
        return BitmapMasks(mask, height=h, width=w)



__all__ = ["show_batch"]

from icevision.utils import *
from icevision.visualize import *


def show_batch(batch, ncols: int = 1, figsize=None, **show_samples_kwargs):
    """Show a single batch from a dataloader.

    # Arguments
        show_samples_kwargs: Check the parameters from `show_samples`
    """
    (tensor_images, *_), records = batch

    for tensor_image, record in zip(tensor_images, records):
        image = tensor_to_image(tensor_image)
        record.set_img(image)

    return show_samples(records, ncols=ncols, figsize=figsize, **show_samples_kwargs)



__all__ = ["show_results", "interp"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.base_show_results import base_show_results
from icevision.models.ultralytics.yolov5.dataloaders import (
    build_infer_batch,
    valid_dl,
    infer_dl,
)
from icevision.models.ultralytics.yolov5.prediction import (
    predict,
    predict_from_dl,
)
from icevision.models.interpretation import Interpretation

from icevision.models.interpretation import _move_to_device
from icevision.core.record_components import LossesRecordComponent
from yolov5.utils.loss import ComputeLoss


def show_results(
    model: nn.Module,
    dataset: Dataset,
    detection_threshold: float = 0.25,
    nms_iou_threshold: float = 0.45,
    num_samples: int = 6,
    ncols: int = 3,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    show: bool = True,
    device: Optional[torch.device] = None,
) -> None:
    return base_show_results(
        predict_fn=predict,
        model=model,
        dataset=dataset,
        num_samples=num_samples,
        ncols=ncols,
        denormalize_fn=denormalize_fn,
        show=show,
        detection_threshold=detection_threshold,
        nms_iou_threshold=nms_iou_threshold,
        device=device,
    )


def loop_yolo(dl, model, losses_stats, device):
    samples_plus_losses = []
    compute_loss = ComputeLoss(model)

    with torch.no_grad():
        for (x, y), sample in pbar(dl):
            torch.manual_seed(0)
            x, y = _move_to_device(x, y, device)
            preds = model(x)
            loss = compute_loss(preds, y)[0]
            loss = {
                "loss_yolo": float(loss.cpu().numpy()),
                "loss_total": float(loss.cpu().numpy()),
            }

            for l in losses_stats.keys():
                losses_stats[l].append(loss[l])

            loss_comp = LossesRecordComponent()
            loss_comp.set_losses(loss)
            sample[0].add_component(loss_comp)
            sample[0].set_img(tensor_to_image(x[0]))
            samples_plus_losses.append(sample[0])
    return samples_plus_losses, losses_stats


_LOSSES_DICT = {
    "loss_yolo": [],
    "loss_total": [],
}

interp = Interpretation(
    losses_dict=_LOSSES_DICT,
    valid_dl=valid_dl,
    infer_dl=infer_dl,
    predict_from_dl=predict_from_dl,
)

interp._loop = loop_yolo
